/*
 * srecord.c
 *
 * Copyright (C) 2009 Texas Instruments Incorporated - http://www.ti.com/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation version 2.
 *
 * This program is distributed "as is" WITHOUT ANY WARRANTY of any
 * kind, whether express or implied; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <bootloader.h>
#include <uart.h>

#define BLOCKLEN 4

#define ENDIAN_SWAP(aa) (((aa&0xFF)<<24)|((aa&0xFF0000)>>8)|((aa&0xFF00)<<8)|((aa&0xFF000000)>>24))

typedef volatile unsigned int	VUint32;
typedef volatile unsigned char	VUint8;

/* Returns the 8 bit Checksum value. This checksum is used
   for the Lookup table(used for CRC-32 checksum) and is
   passed by the HOST.

   for example the string "0000000" returns B0h, because ascii hex value
   of char '0' is 30h and 30+30+30+30+30+30+30=150h-->50h(8 bit)
   B0+50=100-->00(8 bit) checksum=B0
*/
Uint8 Checksum8(Uint32 numBytes, Uint8 *data) {
	Uint8 checksum=0;
	Uint32 i;

	for(i=0;i<numBytes;i++)
		checksum+=data[i];

	return (~checksum+1);
}

/* Assuming only 32 bit (4 byte) aligned memory */
void mem_write(Uint32 *dstAddr, Uint32 *data, Int size)
{
	Int i;

	for (i = 0; i < (size/4); i++ )
	{
		dstAddr[i] = ENDIAN_SWAP(data[i]);
#if 0
		UARTSendInt((Uint32) dstAddr);
		UARTSendData(1, (Uint8 *)"@");
		UARTSendInt(dstAddr[i]);
		UARTSendData(2, (Uint8 *)"\r\n");
#endif
	}

}

static Uint32 GetHexData(Uint8 *src, Uint32 numBytes, Uint32* seq) {
	Uint32 i,j;
	Uint32 temp[8];
	Uint32 numLongs, numAsciiChar, shift;

	if(numBytes == 1)
	{
		numLongs = 1;
		numAsciiChar = 2;
		shift = 4;
	} else if(numBytes == 2) {
		numLongs = 1;
		numAsciiChar = 4;
		shift = 12;
	} else {
		numLongs = numBytes/4;
		numAsciiChar = 8;
		shift = 28;
	}

	for(i=0;i<numLongs;i++) {
		seq[i] = 0;
		for(j=0;j<numAsciiChar;j++) {

			/* Converting ascii to Hex*/
			temp[j] = *src++ - 48;

			if(temp[j] > 22)/* To support lower case a,b,c,d,e,f*/
			   temp[j] = temp[j]-39;
			else if(temp[j]>9)/* To support Upper case A,B,C,D,E,F*/
			   temp[j] = temp[j]-7;

			seq[i] |= (temp[j]<<(shift-(j*4)));
		}
	}
	return E_PASS;
}

Int	validateChecksum(Uint8 *src, Uint8 *checksum)
{
	return 1;
}

Uint32 SFMTDecode(Uint8 *srcAddr, Uint32 size)
{
	VUint32		index;
	VUint8		*hexcharbuf;
	VUint8		*destBuffer;
	Int			cnt;
	Uint32		dstAddr;
	Uint32		data;

	index = 0;
	hexcharbuf = srcAddr;

	/* The first character of an ASCII hex image is always 0x02 */
	if((hexcharbuf[index] != 'S')&&(hexcharbuf[index+1] != '0'))
	{
		return E_FAIL;
	}
	else
    {
		/* Read the size & Ignore that many bytes */
		index += 2;
		GetHexData (&srcAddr[index], 1, (Uint32 *)&cnt);
		index += (cnt * 2) + 3; /* Ignore the count + carriage return */

		while(index <= size)
		{
			hexcharbuf = &srcAddr[index];
			index += 2;

			if(( hexcharbuf[0] == 'S' )&&( hexcharbuf[1] == '3'))
			{
				GetHexData (&srcAddr[index], 1, (Uint32 *)&cnt);
				index += 2;

				GetHexData (&srcAddr[index], 4, &dstAddr);
				index += 8;

				/* Now we might have to add padding if the size is not aligned */
				cnt -= 5;	/* Eliminate the Address 4 Bytes & One for CheckSum */

				while (cnt >= 4)
				{
					hexcharbuf = &srcAddr[index];

					/* Read the whole data. No padding required */
					GetHexData ((Uint8 *)hexcharbuf, 4, &data);
					index += 8;

					/* Write the byte */
					mem_write((Uint32 *)dstAddr, &data, BLOCKLEN);

					/* Increment the address */
					dstAddr = dstAddr + 4;

					/* Reduce the count */
					cnt -= 4;
				}

				/* Now lets see what we need to pad */
				while (cnt > 0)
				{
					destBuffer = (Uint8 *) dstAddr;

					GetHexData (&srcAddr[index], 1, (Uint32 *)&data);
					index += 2;

					*destBuffer++ = data & 0xFF;

					cnt --;
				}

			}

			hexcharbuf = &srcAddr[index];
			/* This is the last character of an ASCII hex image */
			if (( hexcharbuf[0] == 'S' )&&( hexcharbuf[1] == '7' ))
			{
				index += 12;
				return E_PASS;
			}

			/* Ignore all spaces and returns */
			if( (hexcharbuf[0] == ' ')  ||
				(hexcharbuf[0] == '\n') ||
				(hexcharbuf[0] == ',')  ||
				(hexcharbuf[0] == '\r') )
			{
				/**** Do nothing ****/
				index ++;
			}
		}
	}

	return E_PASS;
}

